/*
 * JSyntaxDocument.java
 *
 * Created on 28 ottobre 2007, 22.35
 *
 */

package jsyntax;

import java.awt.Color;
import javax.swing.JTextPane;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.DefaultStyledDocument;
import javax.swing.text.SimpleAttributeSet;
import javax.swing.text.StyleConstants;

import jsyntax.undo.JSyntaxInsertTextAction;
import jsyntax.undo.JSyntaxRemoveTextAction;
import jsyntax.undo.JSyntaxUndoManager;

/**
 * This class implements a code-formatted document, with highlighting for numbers and keywords. Strings and 
 * comments highlighting are not supported at the moment. 
 *
 * @author Dario Lucia
 * @version 1.0
 * @see DefaultStyledDocument, AbstractDocument
 */
public class JSyntaxDocument extends DefaultStyledDocument {
   
    
    private JSyntaxKeywordManager keywords;
    private JSyntaxTokenDivisorManager divisors;
    private JSyntaxTokenizer tokenizer;
    private JTextPane parent;
    
    private JSyntaxUndoManager undoManager;
    
    
    /** Creates a new instance of JSyntaxDocument, with a given JSyntaxTokenDivisorManager and 
     * a given JSyntaxKeywordManager. UndoManager disabled by default.
     *
     * @param  divs the token divisor manager
     * @param  keys the keywords manager
     */
    public JSyntaxDocument(JSyntaxTokenDivisorManager divs, JSyntaxKeywordManager keys) {
        this(divs,keys,false);
    }
    
    /** Creates a new instance of JSyntaxDocument, with a given JSyntaxTokenDivisorManager and 
     * a given JSyntaxKeywordManager. Ask for an undo/redo system support.
     *
     * @param  divs the token divisor manager
     * @param  keys the keywords manager
     * @param undoSupport need for undo/redo subsystem
     */
    public JSyntaxDocument(JSyntaxTokenDivisorManager divs, JSyntaxKeywordManager keys, boolean undoSupport) {
        this.divisors = divs;
        this.keywords = keys;
        this.tokenizer = new JSyntaxTokenizer(divs);
        
        if(undoSupport)
            this.undoManager = new JSyntaxUndoManager(this);
        else
            this.undoManager = null;
    }

    /** Insert a string in the given index, with the given AttributeSet, need for format.
     *  This method doesn't need to be called directly. It's somewhat like a callback called automagically when 
     *  an input action is performed in the textpane.
     *  
     *  @param i index position
     *  @param  string  the text to be added
     *  @param attributeSet the set of properties needed for formatting
     *  @override 
     */
    public void insertString(int i, String string, AttributeSet attributeSet) throws BadLocationException {
        super.insertString(i,string,attributeSet);
        int saveCaret = this.parent.getCaretPosition();
        processDocument(i,i + string.length());
        this.parent.setCaretPosition(saveCaret);
        if(this.undoManager != null)
            this.undoManager.addNewAction(new JSyntaxInsertTextAction(string,i));
    }
    
    
    public void remove(int off, int len) throws BadLocationException {
        // Catch text
        String string = this.getText(off,len);
        super.remove(off,len);
        int saveCaret = this.parent.getCaretPosition();
        processDocument(off, off + len);
        this.parent.setCaretPosition(saveCaret);
        if(this.undoManager != null)
            this.undoManager.addNewAction(new JSyntaxRemoveTextAction(string,off));
    }
    
    private void processDocument(int start,int end) throws BadLocationException {        
        // Get text and document
        String text = this.getText(0,this.getLength());
        // Reset tokenizer
        this.tokenizer.initTokenizer(text, start, end);
        JSyntaxToken token = new JSyntaxToken();
        while(this.tokenizer.hasNext()) {
            this.tokenizer.getNext(token); 
            if(this.keywords.isKeyword(token)) {
                // Keyword format
                // super.remove(token.getIndex(), token.getLenght());
                // super.insertString(token.getIndex(), token.getToken(), this.keywords.getFormat(token.getToken()));                  
                setCharacterAttributes(token.getIndex(),token.getLenght(),this.keywords.getFormat(token.getToken()),true);
            } else if(isNumber(token.getToken())){
                // Number format
                // super.remove(token.getIndex(), token.getLenght());
                // super.insertString(token.getIndex(), token.getToken(), this.keywords.getNumberFormat());  
                setCharacterAttributes(token.getIndex(),token.getLenght(),this.keywords.getNumberFormat(),true);
            } else {
                // Simple text
                // super.remove(token.getIndex(), token.getLenght());
                // super.insertString(token.getIndex(), token.getToken(),this.keywords.getNormalFormat());
                setCharacterAttributes(token.getIndex(),token.getLenght(),this.keywords.getNormalFormat(),true);
            }
        }
        
    }

    private boolean isNumber(String string) {
        try {
            Double.parseDouble(string);
            return true;
        } catch (NumberFormatException e) {
            return false;
        }
    }

    void registerTextPane(JTextPane jSyntaxTextPane) {
        this.parent = jSyntaxTextPane;
    }
    
    
    public void undoAction() {
        if(this.undoManager != null) {
            if(this.undoManager.canUndo())
                this.undoManager.undo();
        }
    }
    
    public void redoAction() {
        if(this.undoManager != null) {
            if(this.undoManager.canRedo())
                this.undoManager.redo();
        }
    }
    
    public void insertStringFromUndo(int i, String string, AttributeSet attributeSet) throws BadLocationException {
        super.insertString(i,string,attributeSet);
        processDocument(i,i + string.length());
        this.parent.setCaretPosition(i + string.length());
    }
    
    
    public void removeFromUndo(int off, int len) throws BadLocationException {
        super.remove(off,len);
        processDocument(off, off + len);
        this.parent.setCaretPosition(off);
    }
    
}
